/* cw.c */

/* RGB color widget window for colored cross slices */

/* VIS-5D version 3.0 */

/*
vis5d program for visualizing five dimensional gridded data sets
Copyright (C) 1990, 1991, 1992  Bill Hibbard, Brian Paul, Dave Santek,
and Andre Battaiola.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 1, or (at your option)
any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/



#include <stdio.h>
#include <math.h>
#include "x.h"


static GC  cwgc;
static Display *cwdpy;
static XFontStruct *cwfont;
static int  cwwidth = 380;
static int  cwheight = 200;
static int  cwwedge;
static int  fontheight = 13;
static char varname[20];
static float minval, maxval;
static int  pos = 50;  /* mouse pointer x position */
static float xscale, yscale;
static int depth, flip;
static unsigned long black_pix, white_pix, red_pix, green_pix, blue_pix;



Window create_cw( dpy, title, x, y, width, height )
Display *dpy;
char title[];
int x,y;
int width, height;
{
   Window win;
   Screen *screen;
   int scr;
   XVisualInfo visinfo;
   Visual *visual;
   XColor xcol;
   XSetWindowAttributes  xswa;
   XGCValues gcv;
   XSizeHints hints;
   Colormap map;
   int swa_flags;
   int cwborder = 0;

   cwdpy = dpy;
   scr = DefaultScreen( dpy );
   flip = 0;

   if (XMatchVisualInfo( dpy, scr, 24, TrueColor, &visinfo )) {
      visual = visinfo.visual;
      depth = 24;
      if (visinfo.red_mask==0xff)
	 flip = 1;
   }
   else {
      visual = DefaultVisual( dpy, scr );
      depth = 8;
      /* die("Unable to get TrueColor visual"); */
   }

   map = XCreateColormap( dpy, RootWindow(dpy, scr), visual, AllocNone );

   /* allocate black, white, red, green, blue colors */
   xcol.red = xcol.green = xcol.blue = 0;
   xcol.flags = DoRed | DoGreen | DoBlue;
   if (XAllocColor(dpy, map, &xcol))
      black_pix = xcol.pixel;
   else
      black_pix = 0;

   xcol.red = xcol.green = xcol.blue = 0xffff;
   xcol.flags = DoRed | DoGreen | DoBlue;
   if (XAllocColor(dpy, map, &xcol))
      white_pix = xcol.pixel;
   else
      white_pix = 0;

   xcol.red = 0xffff;  xcol.green = 0;  xcol.blue = 0;
   xcol.flags = DoRed | DoGreen | DoBlue;
   if (XAllocColor(dpy, map, &xcol))
      red_pix = xcol.pixel;
   else
      red_pix = 0;

   xcol.red = 0;  xcol.green = 0xffff;  xcol.blue = 0;
   xcol.flags = DoRed | DoGreen | DoBlue;
   if (XAllocColor(dpy, map, &xcol))
      green_pix = xcol.pixel;
   else
      green_pix = 0;

   xcol.red = 0;  xcol.green = 0;  xcol.blue = 0xffff;
   xcol.flags = DoRed | DoGreen | DoBlue;
   if (XAllocColor(dpy, map, &xcol))
      blue_pix = xcol.pixel;
   else
      blue_pix = 0;

   /* set window attributes */

   xswa.colormap = map;
   xswa.event_mask = ExposureMask | ButtonMotionMask | KeyPressMask
                | ButtonPressMask | ButtonReleaseMask | StructureNotifyMask;
   xswa.background_pixel = black_pix;
   xswa.border_pixel = white_pix;
   swa_flags = CWColormap | CWEventMask | CWBackPixel | CWBorderPixel;

   /* Create the window */
   win = XCreateWindow( dpy, RootWindow(dpy, scr), x, y,
                        cwwidth, cwheight, cwborder,
		        depth, InputOutput, visual, swa_flags, &xswa);

   if (!win)
      die("Unable to open color window");

   /* Set hints for window manager */
   hints.x = x;
   hints.y = y;
   hints.width  = width;
   hints.height = height;
   hints.min_width = 80;
   hints.min_height = 40;
   hints.flags = USPosition | USSize| PMinSize;
   XSetNormalHints(dpy, win, &hints);
   XSetStandardProperties( dpy, win, title, title, None, NULL, 0, &hints);

   gcv.foreground = white_pix;
   gcv.background = black_pix;
   cwgc = XCreateGC(dpy, win, GCForeground | GCBackground, &gcv);
   if (cwfont = XLoadQueryFont( dpy, "8x13" ) )
      XSetFont( dpy, cwgc, cwfont->fid );

   /* set pointer shape */
   XDefineCursor( dpy, win, XCreateFontCursor(dpy,XC_top_left_arrow) );

   return win;
}





void map_cw( win, min, var, max )
Window win;
float min,max;
char var[];
{

   minval = min;
   maxval = max;
   strcpy(varname, var );
   XMapWindow( cwdpy, win );

}




#define TABLE_SIZE 255

#define R(ctable, entry)  ((ctable[entry] >> 16) & 0xff)
#define G(ctable, entry)  ((ctable[entry] >> 8) & 0xff)
#define B(ctable, entry)  ((ctable[entry]) & 0xff)
#ifdef stellar
#   define MAKECOLOR(r,g,b,a) ( r<<16 | g<<8 | b )
#endif
#if defined(sgi) || defined(ibm)
#   define MAKECOLOR(r,g,b,a) ( a<<24 | r<<16 | g<<8 | b )
#endif



/*** init_ctable ******************************************************
   Initialize a color table.
   Input:  ctable - pointer to the table
           trans - transparency value in [0,255].
**********************************************************************/
init_ctable( ctable, trans )
int ctable[TABLE_SIZE];
int trans;
{
   int i;
   int r,g,b;
   float t;

   for (i=0;i<TABLE_SIZE;i++) {
      t = (float) i / TABLE_SIZE - 0.5;  /* t in [-0.5,0.5) */
      r = 128.0 + 127.0 * atan( 7.0*t ) / 1.57;
      g = 128.0 + 127.0 * cos( t*6.28 );
      b = 128.0 + 127.0 * atan( -7.0*t ) / 1.57;
      ctable[i] = MAKECOLOR(r,g,b,trans);
   }

}



/*** cw_set_trans *****************************************************
   Set the transparency value for all entries in a color table.
   Input:  ctable - pointer to the color table.
           trans - transparency value in [0,255].
**********************************************************************/
cw_set_trans( ctable, trans )
int ctable[TABLE_SIZE];
int trans;
{
   int i, c;

   trans = trans << 24;
   for (i=0;i<TABLE_SIZE;i++) {
#ifdef LEAVEOUT
      c = ctable[i]&0x00ffffff;
      if (c)
	 ctable[i] = c | trans;
      else
	 ctable[i] = 0x0;
#else
      ctable[i] = (ctable[i]&0x00ffffff) | trans;
#endif
   }
}



#define BORD 10



redraw_cw( cwwin, ctable )
Window cwwin;
int ctable[TABLE_SIZE];
{
   int i;
   float x, y, px, py;
   Window gr;
   int gx, gy, gbw, gd;

   /* get window size */
   XGetGeometry(cwdpy, cwwin, &gr, &gx, &gy, &cwwidth, &cwheight, &gbw, &gd);
   cwwedge = cwheight - 2*BORD - 2*fontheight - 1 - 10;

   xscale = (float) (cwwidth-BORD*2) / (float) (TABLE_SIZE-1);
   yscale = (float) (cwwedge-BORD-4) / 255.0;

   /* draw box around color graph */
   XSetForeground( cwdpy, cwgc, white_pix );
   XDrawRectangle( cwdpy, cwwin, cwgc, BORD-1,BORD-1,
                     cwwidth-2*BORD+1, cwwedge-BORD );

   redraw_range( cwwin, ctable, 0, TABLE_SIZE-1 );
   redraw_marker( cwwin );

   LUI_FrameWindow( cwwin, cwwidth, cwheight, 6 );
}




static redraw_range( cwwin, ctable, a,b )
Window cwwin;
int ctable[TABLE_SIZE];
int a,b;
{
   int i;
   int x,y, px,py;
   int x1,y1,x2,y2;
   unsigned long c;

   if (a<0)  a = 0;
   if (b>=TABLE_SIZE)  b = TABLE_SIZE-1;

   /* calc region to update */
   x1 = BORD + (float) a * xscale;
   x2 = BORD + (float) b * xscale;
   if (x2==cwwidth-BORD) x2--;

   /* erase region */
   XSetForeground( cwdpy, cwgc, black_pix );
   XFillRectangle( cwdpy, cwwin, cwgc, x1,BORD, x2-x1+1, cwwedge-BORD-2 );

   /* redraw region of entries in interval [a,b] */
   if (a>0) a--;
   if (b<TABLE_SIZE-1)  b++;

   /* draw red levels */
   XSetForeground( cwdpy, cwgc, red_pix );
   for (i=a;i<=b;i++) {
      x = BORD + (float) i * xscale;
      if (x==cwwidth-BORD) x--;
      y = cwwedge - 3 - (float) R(ctable,i) * yscale;
      if (i!=a)
         XDrawLine( cwdpy, cwwin, cwgc, px,py, x,y );
      px = x;  py = y;
   }

   /* draw green levels */
   XSetForeground( cwdpy, cwgc, green_pix );
   for (i=a;i<=b;i++) {
      x = BORD + (float) i * xscale;
      if (x==cwwidth-BORD) x--;
      y = cwwedge - 3 - (float) G(ctable,i) * yscale;
      if (i!=a)
         XDrawLine( cwdpy, cwwin, cwgc, px,py, x,y );
      px = x;  py = y;
   }

   /* draw blue levels */
   XSetForeground( cwdpy, cwgc, blue_pix );
   for (i=a;i<=b;i++) {
      x = BORD + (float) i * xscale;
      if (x==cwwidth-BORD) x--;
      y = cwwedge - 3 - (float) B(ctable,i) * yscale;
      if (i!=a)
         XDrawLine( cwdpy, cwwin, cwgc, px,py, x,y );
      px = x;  py = y;
   }

   /* draw color wedge */
   if (depth==24) {
      for (i=x1;i<=x2;i++) {
	 x = (float)(i-BORD) * (float)TABLE_SIZE / (float)(cwwidth-BORD*2);
	 c = ctable[x];
	 if (flip)
            c = (c&0xff00ff00) | ((c>>16) & 0xff) | ((c&0xff) <<16);
	 XSetForeground( cwdpy, cwgc, c );
	 XDrawLine( cwdpy, cwwin, cwgc, i,cwwedge+1, i, cwwedge+BORD+5 );
      }
   }

}




static redraw_marker( cwwin )
Window cwwin;
{
   int x;
   char str[30];
   int dir,ascent, descent;
   XCharStruct overall;
   int xpos;
   float val;

   XSetForeground( cwdpy, cwgc, black_pix );
   XFillRectangle( cwdpy, cwwin, cwgc, 6, cwwedge+BORD+6,
                                cwwidth-12, cwheight-cwwedge-BORD-12 );

   /* draw marker below color wedge at ctable[pos] */
   x = BORD + (float) pos * xscale - 1;
   XSetForeground( cwdpy, cwgc, white_pix );
   XDrawLine( cwdpy, cwwin, cwgc, x, cwwedge+BORD+6, x, cwwedge+BORD+20 );
   XDrawLine( cwdpy, cwwin, cwgc, x, cwwedge+BORD+6, x-3, cwwedge+BORD+12 );
   XDrawLine( cwdpy, cwwin, cwgc, x, cwwedge+BORD+6, x+3, cwwedge+BORD+12 );

   /* draw min value */
   sprintf(str,"%.2f",minval );
   XDrawString( cwdpy, cwwin, cwgc, BORD, cwheight-10, str, strlen(str) );

   /* draw variable name and marker value */
   val = minval + (maxval-minval)*( (float) pos / (float) (TABLE_SIZE-1));
   sprintf(str,"%s= %.2f", varname, val );
   XTextExtents( cwfont, str, strlen(str), &dir,&ascent,&descent,&overall );
   xpos = (cwwidth - overall.width) / 2;
   XDrawString( cwdpy, cwwin, cwgc, xpos, cwheight-10, str, strlen(str) );

   /* draw max value */
   sprintf(str,"%.2f",maxval);
   XTextExtents( cwfont, str, strlen(str), &dir,&ascent,&descent,&overall );
   xpos = cwwidth - overall.width - BORD;
   XDrawString( cwdpy, cwwin, cwgc, xpos, cwheight-10, str, strlen(str) );

}





/*** cw_event *********************************************************
   This function handles events in the color widget window.
   Input:  event - the X event
           ctable - pointer to color table to modify
	   trans - current transparency of color table in [0,255].
   Return: 1 - if color table was changed.
           2 - if reset key was pressed
           0 - if no change
**********************************************************************/
int cw_event( event, ctable, trans )
XEvent event;
int ctable[TABLE_SIZE];
int trans;
{
   int entry, value, redraw;
   static int p1 = 0, p2 = 0, p3 = 0;
   static int pentry;
   int i, a, b, newpress;
   int red, green, blue;
   static int move_marker;

   if (event.type==KeyPress) {
      /* check for reset */
      char keybuf[50];
      KeySym key;
      XComposeStatus compose;
      int count;
      count = XLookupString( &event, keybuf, 50, &key, &compose );
      if (count==1 && (keybuf[0]=='r' || keybuf[0]=='R')) {
	 /*
	   init_ctable( ctable, trans );
	   redraw_cw( event.xany.window, ctable );
	 */
	 redraw = 2;
      }
   }

   newpress = 0;
   if (event.type==ButtonPress ) {
      if (event.xbutton.y<cwwedge) {
	 /* change color function */
	 move_marker = 0;
      }
      else {
	 /* change marker position */
	 move_marker = 1;
      }
      if (event.xbutton.button==Button1)  p1 = 1;
      if (event.xbutton.button==Button2)  p2 = 1;
      if (event.xbutton.button==Button3)  p3 = 1;
      pentry = (event.xbutton.x - BORD) * (TABLE_SIZE-1.0)
	/ (float) (cwwidth - BORD*2);
      if (pentry>=TABLE_SIZE)  pentry = TABLE_SIZE-1;
      if (pentry<0)  pentry = 0;
      newpress = 1;
   }
   else if (event.type==ButtonRelease) {
      if (event.xbutton.button==Button1)  p1 = 0;
      if (event.xbutton.button==Button2)  p2 = 0;
      if (event.xbutton.button==Button3)  p3 = 0;
   }

   if ((event.type==MotionNotify || newpress) && (p1 || p2 || p3)) {
      /* calculate which entry in color table to change */
      entry = (event.xbutton.x-BORD) * (TABLE_SIZE-1.0)
              / (float) (cwwidth-BORD*2);
      if (entry>=TABLE_SIZE) entry = TABLE_SIZE-1;
      if (entry<0) entry = 0;
      /* update */
      if (move_marker) {
         /* changing marker position */
         pos = entry;
         redraw_marker( event.xany.window );
      }
      else {
         /* changing color graph */
         value = (cwwedge-3-event.xbutton.y)*255/(float) (cwwedge-BORD-2);
         if (value>255) value = 255;
         if (value<0)  value = 0;

         if (pentry<=entry)  { a = pentry;  b = entry; }
         else  { a = entry;  b = pentry; }

         redraw = 0;

         /* update entries from 'pentry' to 'entry' */
         for (i=a; i<=b; i++) {
	    red = R(ctable,i);
	    green = G(ctable,i);
	    blue = B(ctable,i);
            if (p1) {
               /* modify red */
	       red = value;
               redraw = 1;
            }
            if (p2) {
               /* modify green */
	       green = value;
               redraw = 1;
            }
            if (p3) {
               /* modify blue */
	       blue = value;
	       redraw = 1;
            }
	    ctable[i] = MAKECOLOR(red,green,blue,trans);
         }

         if (redraw) {
            if (pentry<entry)
               redraw_range( event.xany.window, ctable, pentry-1, entry+1 );
            else
               redraw_range( event.xany.window, ctable, entry-1, pentry+1 );
	 }
         pentry = entry;
      }
   }
   else if (event.type==Expose && event.xexpose.count==0) {
      redraw_cw( event.xany.window, ctable );
      redraw = 1;
   }

   return redraw;
}




